package com.ijchart.xychart.render.xy ;

import java.awt.BasicStroke ;
import java.awt.Color ;
import java.awt.Paint ;
import java.awt.geom.Line2D ;
import java.awt.geom.Rectangle2D ;
import java.io.IOException ;
import java.io.ObjectInputStream ;
import java.io.ObjectOutputStream ;
import java.io.Serializable ;

import com.ijchart.xychart.axis.value.AbstractValueAxis ;
import com.ijchart.xychart.chartinfo.RenderShapeCollection ;
import com.ijchart.xychart.constant.ChartOrientation ;
import com.ijchart.xychart.constant.LegendType ;
import com.ijchart.xychart.data.OHLCChartData ;
import com.ijchart.xychart.dataset.IMatrixDataset ;
import com.ijchart.xychart.dataset.SingleRowDataset ;
import com.ijchart.xychart.shape.IChartShape ;
import com.ijchart.xychart.shape.StockChartShape ;
import com.ijchart.xychart.ui.RectangleEdge ;
import com.ijchart.xychart.util.EqualsUtil ;
import com.ijchart.xychart.util.HashcodeUtil ;
import com.ijchart.xychart.util.SerialUtil ;

/**
 * <p>Title: IJChart</p>
 *
 * <p>Description: a chart library for the Java(tm) platform.</p>
 * <zh>ƹƱͼ.</zh>
 * <en>Draw stock shapes.</en>
 *
 * <p>Copyright: Copyright (c) 2013</p>
 *
 * <p>Company: </p>
 *
 * @author zhang feng min,shanghai china.
 * <a href="mailto:ijchartsh@gmail.com">Contract me by email.</a>
 * @version 1.0
 */
public class StockRender extends AbstractXYRender implements Cloneable,
  Serializable {
  /**
   * <zh>л汾ID.</zh>
   * <en>For serialization.</en>
   */
  private static final long serialVersionUID = -1783389990451031444L ;

  /**
   *
   * <zh>volumeͼεȱʡpaint.</zh>
   * <en>The default paint used to fill the volume bars.</en>
   */
  public static final transient Paint DEFAULT_VOLUME_PAINT = Color.gray ;

  /**
   * <zh>۸closeֵopenֵʱ,ͼʹõȱʡpaint.</zh>
   * <en>The default paint used to fill the shape when the price moved up from open to close.</en>
   */
  public static final transient Paint DEFAULT_UP_PAINT = Color.green ;

  /**
   * <zh>۸openֵcloseֵʱ,ͼʹõȱʡpaint</zh>
   * <en>The default paint used to fill the shape when the price moved down from open to close.</en>
   */
  public static final transient Paint DEFAULT_DOWN_PAINT = Color.red ;

  /**
   * <zh>Զʱӵȱʡֵ.</zh>
   * <en>The default factor for calculate shape width.</en>
   */
  public static final double DEFAULT_AUTO_WIDTH_FACTOR = 4.0 / 7 ;

  /**
   * <zh>ȱʡͼ,İٷֱȼ.</zh>
   * <en>The default maximum width of the graphic to percentage of axis length.</en>
   */
  public static final double DEFAULT_MAX_SHAPE_WIDTH_IN_PERCENT = 0.1D ;

  /**
   * <zh>Զʱ.</zh>
   * <en>The factor for calculate shape width.</en>
   */
  private double autoWidthFactor = 0.0 ;

  /**
   * <zh>ͼ֮Сļ.</zh>
   * <en>The minimum gap between one shape and the next.</en>
   */
  private double autoWidthGap = 0.0 ;

  /**
   * <zh>ͼεĿ.</zh>
   * <en>The shape width.</en>
   */
  private double shapeWidth ;

  /**
   * <zh>۸closeֵopenֵʱ,ͼõpaint.</zh>
   * <en>The paint used to fill the shape when the price moved up from open to close.</en>
   */
  private transient Paint upPaint ;

  /**
   * <zh>۸openֵcloseֵʱ,ͼõpaint</zh>
   * <en>The paint used to fill the shape when the price moved down from open to close.</en>
   */
  private transient Paint downPaint ;

  /**
   * <zh>Ƿvolumeͼ.</zh>
   * <en>A flag controlling whether or not volume bars are drawn on the chart. </en>
   */
  private boolean drawVolume ;

  /**
   *
   * <zh>volumeͼεpaint.</zh>
   * <en>The paint used to fill the volume bars (if they are visible).</en>
   */
  private transient Paint volumePaint ;

  /**
   * <zh>ͼε,İٷֱȼ.</zh>
   * <en>The default maximum width of the graphic to percentage of axis length.</en>
   */
  private double maxShapeWidthInPercent ;

  /**
   * <zh>.</zh>
   * <en>The constructor.</en>
   *
   * @param dataset SingleRowDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   */
  public StockRender(SingleRowDataset dataset) {
    this(dataset, -1.0) ;
  }

  /**
   * <zh>.</zh>
   * <en>The constructor.</en>
   *
   * @param dataset SingleRowDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   * @param shapeWidth double <zh>ͼο.</zh>
   * <en>The shape width.</en>
   */
  public StockRender(SingleRowDataset dataset, double shapeWidth) {
    this(dataset, shapeWidth, true) ;
  }

  /**
   * <zh>.</zh>
   * <en>The constructor.</en>
   *
   * @param dataset SingleRowDataset <zh>ݼ (Ϊnull).</zh>
   * <en>The dataset (null not permitted).</en>
   * @param shapeWidth double <zh>ͼο.</zh>
   * <en>The shape width.</en>
   * @param drawVolume double <zh>Ƿvolumeͼ.</zh>
   * <en>A flag indicating whether or not volume bars should be drawn.</en>
   */
  public StockRender(SingleRowDataset dataset, double shapeWidth, boolean drawVolume) {
    super(dataset) ;

    this.shapeWidth = shapeWidth ;
    this.drawVolume = drawVolume ;
    this.volumePaint = DEFAULT_VOLUME_PAINT ;
    this.upPaint = DEFAULT_UP_PAINT ;
    this.downPaint = DEFAULT_DOWN_PAINT ;
    this.autoWidthFactor = DEFAULT_AUTO_WIDTH_FACTOR ;
    this.maxShapeWidthInPercent = DEFAULT_MAX_SHAPE_WIDTH_IN_PERCENT ;

    setBaseItemBorderPaint(new Color(255, 85, 85)) ;
    setBaseItemBorderStroke(new BasicStroke(0.5F)) ;
  }

  /**
   * <zh>ͼεĿ.</zh>
   * <en>Returns the shape width.</en>
   *
   * @return double <zh>ͼεĿ.</zh>
   * <en>The shape width.</en>
   */
  public double getShapeWidth() {
    return this.shapeWidth ;
  }

  /**
   * <zh>ͼεĿ,м{@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the shape width and
   *  sends a {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param width double <zh>ͼεĿ.</zh>
   * <en>The shape width.</en>
   */
  public void setShapeWidth(double width) {
    if (this.shapeWidth != width) {
      this.shapeWidth = width ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>Զʱ.</zh>
   * <en>Returns the factor for calculate shape width.</en>
   *
   * @return double <zh>.</zh>
   * <en>The width factor (generally between 0.0 and 1.0).</en>
   */
  public double getAutoWidthFactor() {
    return this.autoWidthFactor ;
  }

  /**
   *  <zh>Զʱ.</zh>
   * <en>Sets the factor for calculate shape width.</en>
   *
   * @param factor double <zh>,ڱ[0.0,1.0].</zh>
   * <en>The width factor (generally between 0.0 and 1.0).</en>
   */
  public void setAutoWidthFactor(double factor) {
    if (factor < 0 || factor > 1.0) {
      throw new IllegalArgumentException("Argument 'factor' must be in [0.0,1.0]") ;
    }

    if (this.autoWidthFactor != factor) {
      this.autoWidthFactor = factor ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ͼ֮Сļ.</zh>
   * <en>Returns the minimum gap between one shape and the next.</en>
   *
   * @return double <zh>.</zh>
   * <en>The gap.</en>
   */
  public double getAutoWidthGap() {
    return this.autoWidthGap ;
  }

  /**
   * <zh>ͼ֮Сļ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the minimum gap between one shape and the next and
   *  sends a {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param autoWidthGap double <zh>.</zh>
   * <en>The gap.</en>
   */
  public void setAutoWidthGap(double autoWidthGap) {
    if (this.autoWidthGap != autoWidthGap) {
      this.autoWidthGap = autoWidthGap ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>۸closeֵopenֵʱ,ͼõpaint.</zh>
   * <en>Returns the paint used to fill the shape when the price moved up from open to close.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (possibly null).</en>
   */
  public Paint getUpPaint() {
    return this.upPaint ;
  }

  /**
   * <zh>õ۸closeֵopenֵʱ,ͼõpaint,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the paint used to fill the shape when the price moved up from open to close,and
   *  sends a {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null permitted).</en>
   */
  public void setUpPaint(Paint paint) {
    if (!(this.upPaint == null && paint == null)) {
      this.upPaint = paint ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>ص۸openֵcloseֵʱ,ͼõpaint</zh>
   * <en>Returns the paint used to fill the shape when the price moved down from open to close.</en>
   *
   * @return Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (possibly null).</en>
   */
  public Paint getDownPaint() {
    return this.downPaint ;
  }

  /**
   * <zh>õ۸openֵcloseֵʱ,ͼõpaint,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the paint used to fill the shape when the price moved down from open to close,and
   *  sends a {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null permitted).</en>
   */
  public void setDownPaint(Paint paint) {
    if (!(this.downPaint == null && paint == null)) {
      this.downPaint = paint ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>Ƿvolumeͼ.</zh>
   * <en>Returns whether or not volume bars are drawn on the chart. </en>
   *
   * @return boolean <zh>ֵ.</zh>
   * <en>A boolean.</en>
   */
  public boolean isDrawVolume() {
    return this.drawVolume ;
  }

  /**
   * <zh>Ƿvolumeͼ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the flag controlling whether or not volume bars are drawn on the chart and
   *  sends a {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners. </en>
   *
   * @param flag boolean <zh>ֵ.</zh>
   * <en>The flag.</en>
   */
  public void setDrawVolume(boolean flag) {
    if (this.drawVolume != flag) {
      this.drawVolume = flag ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>volumeͼεpaint.</zh>
   * <en>Returns the paint used to fill the volume bars (if they are visible).</en>
   *
   * @return <zh>Paint (Ϊnull).</zh>
   * <en>The paint (possibly null).</en>
   */
  public Paint getVolumePaint() {
    return this.volumePaint ;
  }

  /**
   * <zh>volumeͼεpaint,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>Sets the paint used to fill the volume bars (if they are visible) and
   *  sends a {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param paint Paint <zh>Paint (Ϊnull).</zh>
   * <en>The paint (null not permitted).</en>
   */
  public void setVolumePaint(Paint paint) {
    if (paint == null) {
      throw new IllegalArgumentException("Null 'paint' argument.") ;
    }

    this.volumePaint = paint ;
    fireRenderChanged() ;
  }

  /**
   * <zh>ͼε,İٷֱȼ.</zh>
   * <en>Returns the default maximum width of the graphic to percentage of axis length.</en>
   *
   * @return double <zh>ٷֱ.</zh>
   * <en>The percent.</en>
   */
  public double getMaxShapeWidthInPercent() {
    return this.maxShapeWidthInPercent ;
  }

  /**
   * <zh>ͼε,İٷֱȼ,м
   * {@link com.ijchart.xychart.event.RenderChangeEvent}¼.</zh>
   * <en>The default maximum width of the graphic to percentage of axis length,and
   *  sends a {@link com.ijchart.xychart.event.RenderChangeEvent} to all registered listeners.</en>
   *
   * @param percent double <zh>ٷֱ.</zh>
   * <en>The percent.</en>
   */
  public void setMaxShapeWidthInPercent(double percent) {
    if (this.maxShapeWidthInPercent != percent) {
      this.maxShapeWidthInPercent = percent ;
      fireRenderChanged() ;
    }
  }

  /**
   * <zh>еͼ.</zh>
   * <en>Draws all data items.</en>
   *
   * @param xAxis AbstractValueAxis <zh>xֵ (Ϊnull).</zh>
   * <en>The x value axis (null not permitted).</en>
   * @param xAxisEdge RectangleEdge <zh>xλ(Ϊnull).</zh>
   * <en>The x axis edge.</en>
   * @param yAxis AbstractValueAxis <zh>yֵ (Ϊnull).</zh>
   * <en>The y value axis (null not permitted).</en>
   * @param yAxisEdge RectangleEdge <zh>yλ (Ϊnull).</zh>
   * <en>The y axis edge (null not permitted).</en>
   * @param dataArea Rectangle2D <zh> (Ϊnull).</zh>
   * <en>The chart data area (null not permitted).</en>
   * @param orientation ChartOrientation <zh>ͼķ (Ϊnull).</zh>
   * <en>The chart orientation (null not permitted).</en>
   * @param renderShapes RenderShapeCollection <zh>renderͼμ (Ϊnull).</zh>
   * <en>Collects information about the render (null not permitted).</en>
   * @param alpha float <zh>alphaֵ,ڱ[0.0,1.0].</zh>
   * <en>The alpha transparency (in the range 0.0f to 1.0f, where 0.0f is fully transparent,
   * and 1.0f is fully opaque).</en>
   */
  public void drawAllItems(AbstractValueAxis xAxis, RectangleEdge xAxisEdge,
                           AbstractValueAxis yAxis, RectangleEdge yAxisEdge,
                           Rectangle2D dataArea, ChartOrientation orientation,
                           RenderShapeCollection renderShapes, float alpha) {
    if (xAxis == null) {
      throw new IllegalArgumentException("Null 'xAxis' argument.") ;
    }
    if (xAxisEdge == null) {
      throw new IllegalArgumentException("Null 'xAxisEdge' argument.") ;
    }
    if (yAxis == null) {
      throw new IllegalArgumentException("Null 'yAxis' argument.") ;
    }
    if (yAxisEdge == null) {
      throw new IllegalArgumentException("Null 'yAxisEdge' argument.") ;
    }
    if (dataArea == null) {
      throw new IllegalArgumentException("Null 'dataArea' argument.") ;
    }
    if (orientation == null) {
      throw new IllegalArgumentException("Null 'orientation' argument.") ;
    }
    if (renderShapes == null) {
      throw new IllegalArgumentException("Null 'renderShapes' argument.") ;
    }
    if (alpha < 0.0f || alpha > 1.0f) {
      throw new IllegalArgumentException("The 'alpha' value must be in the range 0.0f to 1.0f.") ;
    }

    IMatrixDataset dataset = getDataset() ;
    if (dataset.isEmpty()) {
      return ;
    }

    int columns = dataset.getColumnCount() ;
    this.fillPaints = getFillPaints(columns) ;
    this.borderPaints = getBorderPaints(columns) ;
    this.borderStrokes = getBorderStrokes(columns) ;

    boolean isVertical = orientation.equals(ChartOrientation.VERTICAL) ; //ֱ

    //stockͼ
    if (dataset.getRowCount() > 0) {
      double maxVolume = findMaxVolume() ;

      boolean allIsNull = true ;
      int notNullCount = 0 ;
      for (int i = 0 ; i < columns ; i++) {
        Object obj = dataset.getData(0, i) ;
        if (obj != null) {
          allIsNull = false ;
          notNullCount++ ;
        }
      }
      if (allIsNull) {
        return ;
      }

      double drawAreaWidth = isVertical ? dataArea.getWidth() : dataArea.getHeight() ;
      double reallyWidth = this.shapeWidth ; //ʵʵͼο
      double volumeWidth = this.shapeWidth ;
      if (reallyWidth <= 0) {
        double shapeArea = drawAreaWidth * (1 - xAxis.getLowerMargin() - xAxis.getUpperMargin()) ;
        double maxPossibleWidth = shapeArea * this.maxShapeWidthInPercent ; //Ŀܿ
        double stockWidth = shapeArea * this.autoWidthFactor / notNullCount ;
        reallyWidth = Math.min(stockWidth, maxPossibleWidth) ;
        volumeWidth = Math.max(reallyWidth, 1) ;
      }

      double yAxisTop = 0.0D, yAxisBottom = 0.0D ; //yȥϿհ׺¿հ׵λ
      if (yAxis.isInverted()) {
        if (isVertical) {
          yAxisTop = dataArea.getY() + dataArea.getHeight() * yAxis.getLowerMargin() ;
          yAxisBottom = dataArea.getMaxY() - dataArea.getHeight() * yAxis.getUpperMargin() ;
        } else {
          yAxisTop = dataArea.getX() + dataArea.getWidth() * yAxis.getLowerMargin() ;
          yAxisBottom = dataArea.getMaxX() - dataArea.getWidth() * yAxis.getUpperMargin() ;
        }
      } else {
        if (isVertical) {
          yAxisTop = dataArea.getY() + dataArea.getHeight() * yAxis.getUpperMargin() ;
          yAxisBottom = dataArea.getMaxY() - dataArea.getHeight() * yAxis.getLowerMargin() ;
        } else {
          yAxisTop = dataArea.getX() + dataArea.getWidth() * yAxis.getUpperMargin() ;
          yAxisBottom = dataArea.getMaxX() - dataArea.getWidth() * yAxis.getLowerMargin() ;
        }
      }
      double maxVolumeHeight = Math.abs(yAxisBottom - yAxisTop) ;

      int currRowCols = dataset.getColumnCount(0) ;
      for (int i = 0 ; i < currRowCols ; i++) {
        OHLCChartData data = (OHLCChartData) dataset.getData(0, i) ;
        if (data != null) {
          long dateMills = data.getDate().getTime() ;
          double open = data.getOpen() ;
          double close = data.getClose() ;
          double high = data.getHigh() ;
          double low = data.getLow() ;
          double volume = data.getVolume() ;

          if (Double.isNaN(open) || Double.isNaN(close)
              || Double.isNaN(high) || Double.isNaN(low)) {
            renderShapes.addDataItemShape(null) ;
            continue ;
          }

          //draw volume
          double xPos = xAxis.valueToJava2D(dataArea, xAxisEdge, dateMills, false) ;
          Rectangle2D volumeRect = null ;
          if (this.drawVolume && maxVolume != 0.0
              && xPos != AbstractValueAxis.INVALID_COORDINATE_VALUE) {
            double volumePercent = volume / maxVolume ;
            double volumeTopMargin = maxVolumeHeight * (1 - volumePercent) ;
            if (isVertical) {
              if (yAxis.isInverted()) {
                volumeRect = new Rectangle2D.Double(xPos - 0.5 * volumeWidth,
                                                    yAxisTop, volumeWidth,
                                                    maxVolumeHeight - volumeTopMargin) ;
              } else {
                volumeRect = new Rectangle2D.Double(xPos - 0.5 * volumeWidth,
                                                    yAxisTop + volumeTopMargin,
                                                    volumeWidth, maxVolumeHeight - volumeTopMargin) ;
              }
            } else {
              if (yAxis.isInverted()) {
                volumeRect = new Rectangle2D.Double(yAxisTop + volumeTopMargin,
                                                    xPos - 0.5 * volumeWidth,
                                                    maxVolumeHeight - volumeTopMargin, volumeWidth) ;
              } else {
                volumeRect = new Rectangle2D.Double(yAxisTop, xPos - 0.5 * volumeWidth,
                                                    maxVolumeHeight - volumeTopMargin, volumeWidth) ;
              }
            }
          }

          //draw open-close rectangle
          Paint openClosePaint = null ;
          if (close > open) {
            if (this.upPaint != null) {
              openClosePaint = this.upPaint ;
            } else {
              openClosePaint = getItemFillPaint(0, i) ;
            }
          } else {
            if (this.downPaint != null) {
              openClosePaint = this.downPaint ;
            } else {
              openClosePaint = getItemFillPaint(0, i) ;
            }
          }

          double openPos = yAxis.valueToJava2D(dataArea, yAxisEdge, open, false) ;
          double closePos = yAxis.valueToJava2D(dataArea, yAxisEdge, close, false) ;
          if (openPos == AbstractValueAxis.INVALID_COORDINATE_VALUE
              || closePos == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
            renderShapes.addDataItemShape(null) ;
            continue ;
          }

          double yPosMaxOpenClose = Math.max(openPos, closePos) ;
          double yPosMinOpenClose = Math.min(openPos, closePos) ;
          Rectangle2D openCloseRect = null ;
          if (isVertical) {
            openCloseRect = new Rectangle2D.Double(xPos - 0.5 * reallyWidth, yPosMinOpenClose,
                                                   reallyWidth, yPosMaxOpenClose - yPosMinOpenClose) ;
          } else {
            openCloseRect = new Rectangle2D.Double(yPosMinOpenClose, xPos - 0.5 * reallyWidth,
                                                   yPosMaxOpenClose - yPosMinOpenClose, reallyWidth) ;
          }

          //draw high-low line
          double highPos = yAxis.valueToJava2D(dataArea, yAxisEdge, high, false) ;
          double lowPos = yAxis.valueToJava2D(dataArea, yAxisEdge, low, false) ;
          if (highPos == AbstractValueAxis.INVALID_COORDINATE_VALUE
              || lowPos == AbstractValueAxis.INVALID_COORDINATE_VALUE) {
            renderShapes.addDataItemShape(null) ;
            continue ;
          }

          Line2D highLowLine = null ;
          if (isVertical) {
            highLowLine = new Line2D.Double(xPos, highPos, xPos, lowPos) ;
          } else {
            highLowLine = new Line2D.Double(highPos, xPos, lowPos, xPos) ;
          }

          Paint borderPaint = getItemBorderPaint(0, i) ;
          StockChartShape stockShape = new StockChartShape(null, volumeRect, this.volumePaint,
                                                           openCloseRect, borderPaint,
                                                           openClosePaint, highLowLine,
                                                           borderPaint) ;

          IChartShape chartShape = createItemShape(orientation, yAxis, 0, i, openCloseRect,
                                                   fillPaints[i], borderPaints[i],
                                                   borderStrokes[i], alpha, stockShape) ;
          renderShapes.addDataItemShape(chartShape) ;
        } else {
          renderShapes.addDataItemShape(null) ;
        }
      }
    }

  }

  /**
   * <zh>volumeֵ.</zh>
   * <en>Find the max volume.</en>
   *
   * @return double
   */
  private double findMaxVolume() {
    double result = Double.NaN ;

    IMatrixDataset dataset = getDataset() ;
    int rowCount = dataset.getRowCount() ;
    if (rowCount > 0) {
      int cols = dataset.getColumnCount(0) ;
      for (int i = 0 ; i < cols ; i++) {
        OHLCChartData data = (OHLCChartData) dataset.getData(0, i) ;
        if (data == null) {
          continue ;
        }

        double volume = data.getVolume() ;
        if (Double.isNaN(result)) {
          result = volume ;
        } else {
          if (!Double.isNaN(volume)) {
            result = Math.max(result, volume) ;
          }
        }
      }
    }

    return result ;
  }

  /**
   * <zh>ͼ.</zh>
   * <en>Return the legend type.</en>
   *
   * @return LegendType <zh>ͼ.</zh>
   * <en>The legend type.</en>
   */
  public LegendType getLegendType() {
    return LegendType.ROW ;
  }

  /**
   * <zh>ֻͼ,ͼε.</zh>
   * <en>Create a shape with no attribute specified.</en>
   *
   * @param rect Rectangle2D <zh>ͼľ (Ϊnull).</zh>
   * <en>The bar rectangle (null not permitted).</en>
   * @param text String <zh>Ҫʾ (Ϊnull).</zh>
   * <en>The text (null permitted).</en>
   * @param appendParam Object <zh>ӵĲ,ڴݶĶ (Ϊnull).</zh>
   * <en>The append parameter,used for special propose (null permitted).</en>
   *
   * @return IChartShape <zh>ͼ.</zh>
   * <en>The shape.</en>
   */
  protected IChartShape createItemShapeOnly(Rectangle2D rect, String text,
                                            Object appendParam) {
    if (appendParam == null) {
      throw new IllegalArgumentException("Null 'appendParam' argument.") ;
    }

    return (IChartShape) appendParam ;
  }

  /**
   * <zh>Ƿ.</zh>
   * <en>Indicates if some other object is "equal to" this one.</en>
   *
   * @param obj <zh>ҪȽϵĶ</zh>
   * <en>the reference object with which to compare.</en>
   * @return <zh>ȷtrue,򷵻false.</zh>
   * <en>true if this object is the same as the obj argument; false otherwise.</en>
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true ;
    }
    if (!(obj instanceof StockRender)) {
      return false ;
    }
    if (!super.equals(obj)) {
      return false ;
    }

    StockRender that = (StockRender) obj ;
    if (this.shapeWidth != that.shapeWidth) {
      return false ;
    }
    if (!EqualsUtil.equals(this.upPaint, that.upPaint)) {
      return false ;
    }
    if (!EqualsUtil.equals(this.downPaint, that.downPaint)) {
      return false ;
    }
    if (this.drawVolume != that.drawVolume) {
      return false ;
    }
    if (this.autoWidthFactor != that.autoWidthFactor) {
      return false ;
    }
    if (this.autoWidthGap != that.autoWidthGap) {
      return false ;
    }
    if (!EqualsUtil.equals(this.volumePaint, that.volumePaint)) {
      return false ;
    }
    if (this.maxShapeWidthInPercent != that.maxShapeWidthInPercent) {
      return false ;
    }

    return true ;
  }

  /**
   * <zh>hashcodeֵ.</zh>
   * <en>Returns a hash code value for the object.</en>
   *
   * @return int <zh>hashcodeֵ.</zh>
   * <en>A hash code value for this object.</en>
   */
  public int hashCode() {
    double hashcode = super.hashCode() ;
    hashcode = 37 * hashcode + 100 * this.autoWidthFactor ;
    hashcode = 37 * hashcode + 100 * this.autoWidthGap ;
    hashcode = 37 * hashcode + 100 * this.shapeWidth ;
    hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.upPaint) ;
    hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.downPaint) ;
    hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.drawVolume) ;
    hashcode = 37 * hashcode + HashcodeUtil.hashCode(this.volumePaint) ;
    hashcode = 37 * hashcode + 100 * this.maxShapeWidthInPercent ;

    return (int) hashcode ;
  }

  /**
   * <zh>ؿ¡Ķ.</zh>
   * <en>Creates and returns a copy of this object.</en>
   * @return object   a clone of this instance.
   * @exception  CloneNotSupportedException  if the object's class does not
   *               support the Cloneable interface. Subclasses
   *               that override the clone method can also
   *               throw this exception to indicate that an instance cannot
   *               be cloned.
   */
  public Object clone() throws CloneNotSupportedException {
    return super.clone() ;
  }

  /**
   * <zh>toString.</zh>
   * <en>Returns a string representation of the object.</en>
   *
   * @return  a string representation of the object.
   */
  public String toString() {
    return "StockRender" ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the output stream.
   *
   * @throws IOException  if there is an I/O error.
   */
  private void writeObject(ObjectOutputStream stream) throws IOException {
    stream.defaultWriteObject() ;
    SerialUtil.writePaint(stream, this.upPaint) ;
    SerialUtil.writePaint(stream, this.downPaint) ;
    SerialUtil.writePaint(stream, this.volumePaint) ;
  }

  /**
   * Provides serialization support.
   *
   * @param stream  the input stream.
   *
   * @throws IOException  if there is an I/O error.
   * @throws ClassNotFoundException  if there is a classpath problem.
   */
  private void readObject(ObjectInputStream stream) throws IOException, ClassNotFoundException {
    stream.defaultReadObject() ;
    this.upPaint = SerialUtil.readPaint(stream) ;
    this.downPaint = SerialUtil.readPaint(stream) ;
    this.volumePaint = SerialUtil.readPaint(stream) ;
  }
}
